"""Runflow errors."""


class RunflowError(Exception):
    """Base Runflow error."""


class RunflowSyntaxError(RunflowError):
    """The flow spec has a syntax error."""


class RunflowReferenceError(RunflowError):
    """The variable is not declared in the flow spec."""


class RunflowTaskTypeError(RunflowError):
    """The task type is not registered."""


class RunflowTaskError(RunflowError):
    """The task execution is failed."""


class RunflowAcyclicTasksError(RunflowError):
    """The task has circular dependency."""


class RunflowReadOnlyError(RunflowError):
    """Attempt to write something but it is read-only."""



"""Utility functions."""

import asyncio
import importlib
import sys


def run_async(coro):
    """Wrap async function as sync call."""
    if sys.version_info[0] == 3 and sys.version_info[1] < 7:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(coro)
        return
    asyncio.run(coro)


async def to_thread(func, *args, **kwargs):
    """Run sync function in thread."""
    if sys.version_info[0] == 3 and sys.version_info[1] < 9:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: func(*args, **kwargs))

    # pylint: disable=no-member
    return await asyncio.to_thread(func, *args, **kwargs)


def import_module(path):
    """Import a path like `path.to.module:class`."""
    try:
        package_name, module_name = path.split(":")
        result = importlib.import_module(package_name)
        getters = module_name.split(".")
        for getter in getters:
            result = getattr(result, getter)
        return result
    except (AttributeError, ValueError) as err:
        raise ImportError(path) from err



"""
runflow - a tool to define an run your workflows.
"""
__all__ = [
    "Flow",
    "Task",
    "loadflow",
    "runflow",
    "runflow_async",
    "cli",
    "registry",
    "get_task_class",
    "register_task_class",
    "RunflowError",
    "RunflowSyntaxError",
    "RunflowReferenceError",
    "RunflowTaskTypeError",
    "RunflowTaskError",
    "RunflowAcyclicTasksError",
]


try:
    from importlib.metadata import entry_points
except ImportError:  # python < 3.8
    try:
        from importlib_metadata import entry_points
    except ImportError:
        entry_points = None

from . import autoloader  # noqa
from .cli import cli
from .core import Flow, Task
from .errors import (
    RunflowAcyclicTasksError,
    RunflowError,
    RunflowReferenceError,
    RunflowSyntaxError,
    RunflowTaskError,
    RunflowTaskTypeError,
)
from .registry import (
    registry,
    register_task_class,
    get_task_class,
)
from .run import loadflow, runflow, runflow_async


if entry_points is not None:
    try:
        _entry_points = entry_points()
    except TypeError:
        pass  # importlib-metadata < 0.8
    else:
        # Python 3.10+ / importlib_metadata >= 3.9.0
        _tasks = (
            _entry_points.select(group="runflow.tasks")
            if hasattr(_entry_points, "select")
            else _entry_points.get("runflow.tasks", [])
        )
        for _task in _tasks:
            register_task_class(_task.name, _task.value)


if __name__ == "__main__":
    cli()



"""Command-line interface for runflow."""

import argparse
import logging
import sys

from networkx.drawing.nx_agraph import to_agraph

from . import hcl2
from .run import loadflow, runflow


def cli_abort(message):
    """Print abort message."""
    print(message, file=sys.stderr)
    sys.exit(1)


def cli_parser():
    """Parse runflow arguments."""
    parser = argparse.ArgumentParser(
        prog="runflow",
        description="Runflow - a lightweight workflow engine.",
    )
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "CRITICAL"],
        default="INFO",
        help="Logging level",
    )

    subparsers = parser.add_subparsers(help="", dest="subparser_name")

    run_parser = subparsers.add_parser("run", help="Run a Runflow spec file")
    run_parser.add_argument("specfile", help="Path to a Runflow spec file")
    run_parser.add_argument(
        "--var",
        dest="vars",
        action="append",
        help="A variable key-value pair",
    )
    run_parser.add_argument(
        "--var-file",
        dest="varfiles",
        action="append",
        help="Path to a file including variables",
    )

    visualize_parser = subparsers.add_parser(
        "visualize", help="Visualize a Runflow spec file"
    )
    visualize_parser.add_argument(
        "specfile", help="Path to a Runflow spec file"
    )
    visualize_parser.add_argument(
        "--output",
        default="visualize.svg",
        help="The output of the flow graph visualization file.",
    )

    return parser


def cli_parser_var(var):
    """Parse runflow `--var` value."""
    key, value = var.split("=")
    return key.strip(), value.strip()


def _load_specfile(specfile):
    if specfile.endswith(".hcl"):
        return {"path": specfile}

    if specfile == "-":
        return {"source": sys.stdin.read()}

    return {"module": specfile}


def cli_subcommand_run(args):
    """Command: `runflow run`."""
    vars = {}
    for varfile in args.varfiles or []:
        with open(varfile) as file:
            ctx = {}
            for key, value in hcl2.loads(file.read()).items():
                eval_value = hcl2.evaluate(value, ctx)
                ctx[key] = eval_value
                vars[key] = eval_value

    for var in args.vars or []:
        try:
            key, value = cli_parser_var(var)
            vars[key] = value
        except ValueError:
            cli_abort(f"Invalid --var option: {var}")

    loader = _load_specfile(args.specfile)
    runflow(vars=vars, **loader)


def cli_subcommand_visualize(args):
    """Command: `runflow visualize`."""
    flow = loadflow(**_load_specfile(args.specfile))
    agraph = to_agraph(flow.graph)
    agraph.layout("dot")
    agraph.draw(args.output)


SUBCOMMANDS = dict(
    run=cli_subcommand_run,
    visualize=cli_subcommand_visualize,
)


def cli(argv=None):
    """Command: `runflow`."""
    parser = cli_parser()
    args, _rest = parser.parse_known_args(argv)

    logging_format = "[%(asctime)-15s] %(message)s"
    logging.basicConfig(level=args.log_level, format=logging_format)

    if args.subparser_name in SUBCOMMANDS:
        SUBCOMMANDS[args.subparser_name](args)
    else:
        parser.print_help()



from sqlalchemy import create_engine, text


class SqlExecTask:
    def __init__(self, dsn, sql, parameters=None, autocommit=True):
        self.engine = create_engine(str(dsn), echo=False, future=True)
        self.sqls = sql
        self.parameters = parameters or []
        self.autocommit = autocommit

    def run(self, context):
        with self.engine.begin() as conn:
            conn = conn.execution_options(autocommit=self.autocommit)
            for sql in self.sqls:
                statement = text(sql["statement"])
                parameters = sql.get("parameters") or None
                conn.execute(statement, parameters)
        return {}


class SqlRowTask:
    def __init__(self, dsn, sql, parameters=None):
        self.engine = create_engine(str(dsn), echo=False, future=True)

        if len(sql) != 1:
            raise ValueError("sql_row supports only one sql block.")

        self.sql = sql[0]
        self.parameters = parameters or []

    def run(self, context):
        with self.engine.begin() as conn:
            statement = text(self.sql["statement"])
            parameters = self.sql.get("parameters") or None
            result = conn.execute(statement, parameters)
            rows = [dict(m) for m in result.mappings()]
            return {"rows": rows}



"""
Auto-load a .hcl file using Python import string.

Say, we have a file `mypackage/my_flow.hcl`,
you can get the flow object by running the statements like below::

    >>> import runflow.autoloader
    >>> from mypackage.my_flow import flow
"""
import os
import sys
from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_loader

from .core import Flow

EXT_RUNFLOW = ".hcl"


__all__ = []


# pylint: disable=abstract-method
class FlowLoader(Loader):
    """Load .hcl file.

    See https://docs.python.org/3/library/importlib.html#importlib.abc.Loader
    """

    def __init__(self, full_path):
        self._full_path = full_path
        self._flow = None

    def create_module(self, spec):
        """Load .hcl to a Flow object."""
        try:
            self._flow = Flow.from_specfile(self._full_path)
        except Exception as err:
            raise ImportError from err

    def exec_module(self, module):
        """Execute the flow module when the module is loaded.

        This method assigns variable `flow` to the loaded Flow object.
        """
        module.__dict__.update({"flow": self._flow})


# pylint: disable=no-self-use
class FlowMetaPathFinder(MetaPathFinder):
    """A meta path finder for a .hcl file."""

    def find_spec(self, fullname, path, target=None):
        """Find import-related information used to load a Flow module."""
        mod_name = fullname.split(".")[-1]
        paths = path if path else [os.path.abspath(os.curdir)]
        for check_path in paths:
            full_path = os.path.join(check_path, mod_name + EXT_RUNFLOW)
            if os.path.exists(full_path):
                return spec_from_loader(fullname, FlowLoader(full_path))
        return None


sys.meta_path.insert(0, FlowMetaPathFinder())



from base64 import b64decode

import httpx


class HttpRequestTask:
    def __init__(
        self,
        method,
        url,
        params=None,
        headers=None,
        cookies=None,
        content=None,
        b64content=None,
        data=None,
        files=None,
        json=None,
        auth=None,
        allow_redirects=True,
        timeout=None,
        raise_for_status=True,
    ):
        self.method = method
        self.url = url
        self.params = params
        self.headers = headers
        self.cookies = cookies
        self.content = b64decode(b64content) if b64content else content
        self.data = data
        self.files = files
        self.json = json
        self.auth = auth
        self.allow_redirects = allow_redirects
        self.timeout = timeout
        self.raise_for_status = raise_for_status

    async def run(self, context):
        async with httpx.AsyncClient() as client:
            response = await client.request(
                method=self.method,
                url=self.url,
                params=self.params,
                headers=self.headers,
                cookies=self.cookies,
                content=self.content,
                data=self.data,
                files=self.files,
                json=self.json,
                auth=self.auth,
                allow_redirects=self.allow_redirects,
                timeout=self.timeout,
            )
            if self.raise_for_status:
                response.raise_for_status()
            return {"response": response}



"""
This module is a fork of amplify-education/python-hcl2.

Some reasons not using python-hcl2:

* It returns Dict, instead of an AST. For the Runflow's use case, we
  need to populate some fields in the runtime.

* The library seems problematic for a while:
  * It cannot parse exp values: 1e7 => lark.exceptions.UnexpectedToken.
  * It cannot transform full splat:
    a=b[*].1 => {'a': Tree('full_splat_expr_term', ['b', 1])}
  * It turns attribute values to a list: a = 1 => "a": [1].

* I want the parser can import and call Python functions in HCL2.
"""

import itertools
import json
import operator
import re
import textwrap
from datetime import datetime
from functools import singledispatch
from typing import Any, Dict, List, Set, Union

from dateutil.parser import parse as parse_datetime
from lark import Discard, Transformer
from tenacity import (
    wait_fixed,
    wait_random,
    wait_exponential,
    wait_random_exponential,
    wait_chain,
)

from runflow.errors import RunflowReferenceError
from runflow.utils import import_module
from runflow.hcl2_parser import Lark_StandAlone

HEREDOC_PATTERN = re.compile(
    r"<<-?([a-zA-Z][a-zA-Z0-9._-]+)\n((.|\n)*?)\n\s*\1", re.S
)

INTERPOLATION = re.compile(r'\${ *((?:"(?:[^"\\]|\\.)*"|[^}"]+)+) *}')


def parse_template(string):
    result = []
    previous = 0

    for interpolation_match in INTERPOLATION.finditer(string):
        start, end = interpolation_match.span()

        if previous != start:
            result.append(StringLit(string[previous:start]))

        expr = interpolation_match.group(1)
        result.append(loads(expr, start="eval"))
        previous = end

    if not result:
        return StringLit(string)

    if previous != len(string):
        result.append(StringLit(string[previous:]))

    return JoinedStr(result)


class Module(dict):
    pass


class StringLit(str):
    pass


class JoinedStr:
    def __init__(self, elements):
        self.elements = elements

    def __repr__(self):
        return "<JoinedStr>"

    def __eq__(self, o):
        return isinstance(o, JoinedStr) and o.elements == self.elements


class Attribute(dict):
    def __init__(self, key, value):
        self.key = key
        self.value = value
        super().__init__({key: value})

    def merge_to(self, res):
        res = dict(res)
        res.update(self)
        return res


class Block(dict):
    def merge_to(self, res):
        res = dict(res)
        for key, value in self.items():
            key = str(key)

            if key not in res:
                res[key] = [value]
                continue

            if not isinstance(res[key], list):
                res[key] = [res[key], value]
                continue

            if isinstance(value, list):
                res[key].extend(value)
            else:
                res[key].append(value)

        return res


class Interpolation(str):
    def __init__(self, expr):
        super().__init__()
        self.expr = expr
        self.value = "${%s}" % expr

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

    def __eq__(self, o):
        return isinstance(o, Interpolation) and self.expr == o.expr


class Identifier(str):
    pass


class GetIndex:
    def __init__(self, expr, index):
        self.expr = expr
        self.index = index
        self.value = "%s[%s]" % (self.expr, self.index)

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

    def __eq__(self, o):
        return (
            isinstance(o, GetIndex)
            and self.expr == o.expr
            and self.index == o.index
        )

    @property
    def attr_chain(self):
        chain = []
        extract_attr_chain(self, chain)
        return chain


class GetAttr:
    def __init__(self, expr, attr):
        self.expr = expr
        self.attr = attr
        self.value = "%s.%s" % (self.expr, self.attr)

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

    def __eq__(self, o):
        return (
            isinstance(o, GetAttr)
            and self.expr == o.expr
            and self.attr == o.attr
        )

    @property
    def attr_chain(self):
        chain = []
        extract_attr_chain(self, chain)
        return chain


@singledispatch
def extract_attr_chain(ast, chain):
    chain.append(ast)


@extract_attr_chain.register
def _(ast: GetAttr, chain):
    extract_attr_chain(ast.expr, chain)
    chain.append(ast.attr)


@extract_attr_chain.register
def _(ast: GetIndex, chain):
    extract_attr_chain(ast.expr, chain)
    chain.append(ast.index)


class Splat:
    def __init__(self, array, elements):
        self.array = array
        self.elements = elements

    def __repr__(self):
        return "%s.*.%s" % (
            self.array,
            ".".join([str(e) for e in self.elements]),
        )

    def __eq__(self, o):
        return (
            isinstance(o, Splat)
            and self.array == o.array
            and self.elements == o.elements
        )


class Kwargs:
    def __init__(self, args=None):
        self.args = args


class Call:
    def __init__(self, func_name, args, kwargs):
        self.func_name = func_name
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return "%s(%s)" % (self.func_name, ",".join(str(a) for a in self.args))

    def __eq__(self, o):
        return (
            isinstance(o, Call)
            and self.func_name == o.func_name
            and self.args == o.args
        )


class Conditional:
    def __init__(self, predicate, true_expr, false_expr):
        self.predicate = predicate
        self.true_expr = true_expr
        self.false_expr = false_expr

    def __repr__(self):
        return "%s ? %s : %s" % (
            self.predicate,
            self.true_expr,
            self.false_expr,
        )

    def __eq__(self, o):
        return (
            isinstance(o, Conditional)
            and self.predicate == o.predicate
            and self.true_expr == o.true_expr
            and self.false_expr == o.false_expr
        )


class Operation:
    def __init__(self, elements):
        self.elements = elements

    def __repr__(self):
        return "(%s)" % ("".join([str(e) for e in self.elements]))

    def __eq__(self, o):
        return isinstance(o, Operation) and self.elements == o.elements


class ListExpr:
    def __init__(self, element_id, array, element, condition):
        self.element_id = element_id
        self.array = array
        self.element = element
        self.condition = condition

    def __repr__(self):
        return "[for %s in %s: %s%s]" % (
            (
                ",".join(self.element_id)
                if isinstance(self.element_id, tuple)
                else self.element_id
            ),
            self.array,
            self.element,
            f" if {self.condition}" if self.condition else "",
        )

    def __eq__(self, o):
        return (
            isinstance(o, ListExpr)
            and self.element_id == o.element_id
            and self.array == o.array
            and self.element == o.element
            and self.condition == o.condition
        )


# pylint: disable=too-many-arguments
class DictExpr:
    def __init__(self, element_id, array, key, value, condition):
        self.array = array
        self.element_id = element_id
        self.key = key
        self.value = value
        self.condition = condition

    def __repr__(self):
        return "{for %s in %s: %s => %s%s}" % (
            (
                ",".join(self.element_id)
                if isinstance(self.element_id, tuple)
                else self.element_id
            ),
            self.array,
            self.key,
            self.value,
            f" if {self.condition}" if self.condition else "",
        )

    def __eq__(self, o):
        return (
            isinstance(o, DictExpr)
            and self.element_id == o.element_id
            and self.array == o.array
            and self.key == o.key
            and self.value == o.value
            and self.condition == o.condition
        )


class Not:
    def __init__(self, expr):
        self.expr = expr

    def __repr__(self):
        return "(!%s)" % self.expr

    def __eq__(self, o):
        return isinstance(o, Not) and self.expr == o.expr


def strip_new_line_tokens(args: List) -> List:
    return [
        arg for arg in args if arg != "\n" and not isinstance(arg, Discard)
    ]


# pylint: disable=too-many-public-methods,no-self-use
class AstTransformer(Transformer):
    def module(self, args: List) -> Dict:
        args = strip_new_line_tokens(args)
        return Module(args[0])

    def eval(self, args: List) -> Dict:
        return args[0]

    def quoted_template_expr(self, args: Any):
        return parse_template(args[0])

    def string_lit(self, args: Any):
        return StringLit(args[0])

    # pylint: disable=invalid-name
    def STRING_LIT(self, args: Any):
        return self.strip_quotes("".join([str(arg) for arg in args]))

    def _heredoc_template(self, args: List) -> str:
        match = HEREDOC_PATTERN.match(str(args[0]))
        if not match:
            raise RuntimeError("Invalid Heredoc token: %s" % args[0])
        return match.group(2)

    def heredoc_template(self, args: List) -> str:
        return parse_template(self._heredoc_template(args))

    def heredoc_template_trim(self, args: List) -> str:
        return parse_template(textwrap.dedent(self._heredoc_template(args)))

    def float_lit(self, args: List) -> float:
        return float("".join([str(arg) for arg in args]))

    def int_lit(self, args: List) -> int:
        return int("".join([str(arg) for arg in args]))

    def expr_term(self, args: List) -> Any:
        args = strip_new_line_tokens(args)
        if args[0] == "true":
            return True
        if args[0] == "false":
            return False
        if args[0] == "null":
            return None
        if args[0] == "(":
            return args[1]
        return args[0]

    def attribute(self, args: List) -> Attribute:
        key = str(args[0])
        if key.startswith('"') and key.endswith('"'):
            key = key[1:-1]
        value = self.to_string_dollar(args[1])
        return Attribute(key, value)

    def block(self, args: List) -> Block:
        args = strip_new_line_tokens(args)
        if isinstance(args[-1], str):
            args.append({})

        result: Dict[str, Any] = {}
        current_level = result
        for arg in args[0:-2]:
            current_level[self.strip_quotes(arg)] = {}
            current_level = current_level[self.strip_quotes(arg)]

        current_level[self.strip_quotes(args[-2])] = args[-1]
        return Block(result)

    def body(self, args: List) -> Module:
        args = strip_new_line_tokens(args)
        result: Dict[str, Any] = {}
        for arg in args:
            result = arg.merge_to(result)
        return result

    def to_string_dollar(self, value: Any) -> Any:
        if isinstance(
            value,
            (
                GetAttr,
                GetIndex,
                Splat,
                Call,
                Conditional,
                Operation,
                ListExpr,
                DictExpr,
                Not,
                Identifier,
            ),
        ):
            return Interpolation(value)
        if isinstance(value, str):
            if value.startswith('"') and value.endswith('"'):
                return str(value)[1:-1]
            return str(value)
        return value

    def identifier(self, value: Any) -> Identifier:
        return Identifier(str(value[0]))

    def index_expr_term(self, args: List) -> GetIndex:
        args = strip_new_line_tokens(args)
        return GetIndex(args[0], args[1])

    def index(self, args: List) -> Any:
        args = strip_new_line_tokens(args)
        return self.strip_quotes(args[0])

    def get_attr_expr_term(self, args: List) -> GetAttr:
        return GetAttr(args[0], args[1])

    def attr_splat_expr_term(self, args: List) -> Splat:
        return Splat(args[0], args[1])

    def attr_splat(self, args: List):
        return args

    def full_splat_expr_term(self, args: List) -> Splat:
        return Splat(args[0], args[1])

    def full_splat(self, args: List):
        return args

    def tuple(self, args: List) -> List:
        args = strip_new_line_tokens(args)
        return [self.to_string_dollar(arg) for arg in args]

    def object_elem(self, args: List) -> Dict:
        key = self.strip_quotes(args[0])
        value = self.to_string_dollar(args[1])
        return {key: value}

    def object(self, args: List) -> Dict:
        args = strip_new_line_tokens(args)
        result: Dict[str, Any] = {}
        for arg in args:
            result.update(arg)
        return result

    def function_call(self, args: List) -> Call:
        args = strip_new_line_tokens(args)
        func_name = str(args[0])

        if len(args) == 1:
            return Call(func_name, [], {})

        func_args, func_kwargs = args[1]
        return Call(func_name, func_args, func_kwargs)

    def kwarg(self, args: List):
        return Kwargs()

    def arguments(self, args: List) -> List:
        if args and isinstance(args[-1], Kwargs):
            return (args[:-2], args[-2])
        return (args, [])

    def conditional(self, args: List) -> Conditional:
        args = strip_new_line_tokens(args)
        if len(args) == 1:
            return args[0]
        return Conditional(args[0], args[1], args[2])

    def binary_operator(self, args: List):
        return str(args[0])

    def binary_op(self, args: List):
        args = strip_new_line_tokens(args)
        if len(args) == 1:
            return args[0]
        return Operation(args)

    binary_or_op = binary_op
    binary_or_operator = binary_operator

    binary_and_op = binary_op
    binary_and_operator = binary_operator

    binary_eq_op = binary_op
    binary_eq_operator = binary_operator

    binary_test_op = binary_op
    binary_test_operator = binary_operator

    binary_term_op = binary_op
    binary_term_operator = binary_operator

    binary_factor_op = binary_op
    binary_factor_operator = binary_operator

    def for_intro(self, args: List):
        args = strip_new_line_tokens(args)

        if len(args) == 5:
            return [args[1], args[3]]

        if len(args) == 7:
            return [(args[1], args[3]), args[5]]

        raise ValueError(f"invalid for intro: {args}")

    def for_cond(self, args: List):
        args = strip_new_line_tokens(args)
        return args[-1]

    def for_tuple_expr(self, args: List):
        args = strip_new_line_tokens(args)
        element_id, array = args[1]
        element = args[2]
        condition = args[3] if len(args) > 4 else None
        return ListExpr(element_id, array, element, condition)

    def for_object_expr(self, args: List):
        args = strip_new_line_tokens(args)
        element_id, array = args[1]
        key, value = args[2], args[4]
        condition = args[5] if len(args) > 6 else None
        return DictExpr(element_id, array, key, value, condition)

    def unary_op(self, args: List):
        if args[0] == "-":
            return Operation([0, "-", args[1]])

        if args[0] == "!":
            return Not(args[1])

        raise ValueError(f"invalid operator: {args[0]}")

    def new_line_and_or_comma(self, _: List) -> Discard:
        return Discard()

    def new_line_or_comment(self, _: List) -> Discard:
        return Discard()

    def strip_quotes(self, value: Any) -> Any:
        if isinstance(value, str):
            if value.startswith('"') and value.endswith('"'):
                return str(value)[1:-1]
        return value


hcl2 = Lark_StandAlone(transformer=AstTransformer())


def loads(source, start="module"):
    if start == "module":
        source = source + "\n"

    return hcl2.parse(source, start)


@singledispatch
def _for_iterable(iterable, id_count=1):
    raise ValueError("invalid iterable type: {iterable}")


@_for_iterable.register(tuple)
@_for_iterable.register(list)
def _(iterable, id_count=1):
    for index, element in enumerate(iterable):
        yield element if id_count == 1 else (index, element)


@_for_iterable.register(dict)
def _(iterable, id_count=1):
    for key, value in iterable.items():
        yield key if id_count == 1 else (key, value)


@_for_iterable.register(set)
def _(iterable, id_count=1):
    for element in iterable:
        yield element if id_count == 1 else (element, element)


@singledispatch
def resolve_deps(value: Any, deps: Set):
    pass


@resolve_deps.register
def _(value: Interpolation, deps: Set):
    resolve_deps(value.expr, deps)


@resolve_deps.register
def _(value: GetAttr, deps: Set):
    task_keys = list(value.attr_chain)
    if task_keys and task_keys[0] == "task" and len(task_keys) >= 3:
        deps.add(".".join(task_keys[:3]))


@resolve_deps.register
def _(value: JoinedStr, deps: Set):
    for element in value.elements:
        resolve_deps(element, deps)


@resolve_deps.register
def _(value: list, deps: Set):
    for _value in value:
        resolve_deps(_value, deps)


@resolve_deps.register
def _(value: dict, deps: Set):
    for _value in value.values():
        resolve_deps(_value, deps)


@singledispatch
def evaluate(ast, _):
    return ast


@evaluate.register
def _(ast: Module, env):
    return {
        key: evaluate(attribute_or_block, env)
        for key, attribute_or_block in ast.items()
    }


@evaluate.register
def _(ast: Interpolation, env):
    return evaluate(ast.expr, env)


@evaluate.register(GetIndex)
@evaluate.register(GetAttr)
def _(ast: Union[GetIndex, GetAttr], env):
    result = env
    for attr in ast.attr_chain:
        if isinstance(attr, Splat):
            result = evaluate(attr, env)
        elif isinstance(attr, str) and hasattr(result, attr):
            result = getattr(result, attr)
        else:
            try:
                result = result[attr]
            except KeyError as error:
                message = list(ast.attr_chain)
                raise RunflowReferenceError(message) from error
    return result


@evaluate.register
def _(ast: Identifier, env):
    return env[ast]


@evaluate.register
def _(ast: StringLit, _):
    return ast


@evaluate.register
def _(ast: JoinedStr, env):
    result = []
    if len(ast.elements) == 1:
        if isinstance(ast.elements[0], StringLit):
            return str(ast.elements[0])
        return evaluate(ast.elements[0], env)

    for node in ast.elements:
        if isinstance(node, StringLit):
            result.append(str(node))
        else:
            result.append(str(evaluate(node, env)))
    return "".join(result)


@evaluate.register
def _(ast: Not, env):
    return not evaluate(ast.expr, env)


@evaluate.register
def _(ast: Splat, env):
    array = evaluate(ast.array, env)

    def extract(elem, attrs):
        result = elem
        for attr in attrs:
            if isinstance(attr, str) and hasattr(result, attr):
                result = getattr(result, attr)
            else:
                result = result[attr]
        return result

    return [extract(elem, ast.elements) for elem in array]


@evaluate.register
def _(ast: ListExpr, env):
    result = []
    id_count = 2 if isinstance(ast.element_id, tuple) else 1
    for elem in _for_iterable(evaluate(ast.array, env), id_count):
        if id_count == 2:
            index_id, element_id = ast.element_id
            _index, _elem = elem
            newenv = dict(
                env,
                **{
                    str(element_id): _elem,
                    str(index_id): _index,
                },
            )
        else:
            newenv = dict(env, **{str(ast.element_id): elem})
        if ast.condition and not evaluate(ast.condition, newenv):
            continue
        result.append(evaluate(ast.element, newenv))
    return result


@evaluate.register
def _(ast: DictExpr, env):
    result = {}
    id_count = 2 if isinstance(ast.element_id, tuple) else 1
    for elem in _for_iterable(evaluate(ast.array, env), id_count):
        if id_count == 2:
            index_id, element_id = ast.element_id
            _index, _elem = elem
            newenv = dict(
                env,
                **{
                    str(element_id): _elem,
                    str(index_id): _index,
                },
            )
        else:
            newenv = dict(env, **{str(ast.element_id): elem})
        if ast.condition and not evaluate(ast.condition, newenv):
            continue
        key = evaluate(ast.key, newenv)
        value = evaluate(ast.value, newenv)
        result[key] = value
    return result


@evaluate.register
def _(ast: Conditional, env):
    cond = evaluate(ast.predicate, env)

    if cond:
        return evaluate(ast.true_expr, env)

    return evaluate(ast.false_expr, env)


@evaluate.register
def _(ast: Call, env):
    func_name = str(ast.func_name)
    if ":" in func_name:
        func = import_module(func_name)
    elif func_name in __builtins__:
        func = __builtins__[func_name]
    elif func_name in FUNCS:
        func = FUNCS[func_name]
    elif func_name in env.get("func"):
        func = env["func"][func_name]
    else:
        raise NameError(f"function {ast.func_name} is not defined")

    args = evaluate(ast.args, env)
    kwargs = evaluate(ast.kwargs, env)
    if isinstance(kwargs, dict):
        _args = args
        _kwargs = kwargs
    elif isinstance(kwargs, list):
        _args = args + kwargs
        _kwargs = {}
    else:
        _args = args + [kwargs]
        _kwargs = {}

    return func(*_args, **_kwargs)


OPERATORS = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
    "%": operator.mod,
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
    "==": operator.eq,
    "!=": operator.ne,
}


@evaluate.register
def _(ast: Operation, env):
    result = None
    _operator = None
    for index, elem in enumerate(ast.elements):
        if index == 0:
            result = evaluate(elem, env)
        elif index % 2 == 1:
            _operator = elem
        elif _operator in OPERATORS:
            result = OPERATORS[_operator](result, evaluate(elem, env))
        elif _operator == "||":
            result = result or evaluate(elem, env)
        elif _operator == "&&":
            result = result and evaluate(elem, env)
        else:
            raise ValueError(f"invalid operator: {_operator}")

    return result


@evaluate.register
def _(ast: list, env):
    return [evaluate(a, env) for a in ast]


@evaluate.register
def _(ast: dict, env):
    return {k: evaluate(v, env) for k, v in ast.items()}


FUNCS = {
    "lower": lambda s: s.lower(),
    "upper": lambda s: s.upper(),
    "split": lambda sep, s: s.split(sep),
    "join": lambda sep, s: sep.join(s),
    "tojson": json.dumps,
    "concat": lambda *s: list(itertools.chain(*s)),
    "datetime": datetime,
    "todatetime": parse_datetime,
    "call": lambda obj, meth, args=None, kwargs=None: (
        getattr(obj, meth)(*(args or []), **(kwargs or {}))
    ),
    "wait_fixed": wait_fixed,
    "wait_random": wait_random,
    "wait_exponential": wait_exponential,
    "wait_random_exponential": wait_random_exponential,
    "wait_chain": lambda s: wait_chain(*s),
}



"""Core module for runflow."""

import enum
import inspect
import logging
import traceback
from functools import reduce

import networkx
from decouple import config
from tenacity import (
    after_log,
    retry,
    stop_after_delay,
    stop_after_attempt,
)

from . import hcl2, utils
from .errors import (
    RunflowAcyclicTasksError,
    RunflowSyntaxError,
)
from .hcl2_parser import LarkError
from .registry import get_task_class, register_task_class

logger = logging.getLogger(__name__)


DEPENDS_ON_KEY = "_depends_on"


class TaskStatus(enum.Enum):
    """Task execution status."""

    PENDING = enum.auto()
    SUCCESS = enum.auto()
    FAILED = enum.auto()
    CANCELED = enum.auto()


class TaskResult:
    """Task execution result."""

    def __init__(self, status):
        self.status = status
        self._result = None
        self._exception = None

    @property
    def result(self):
        """Get task result."""
        if self._exception:
            raise ValueError(
                "Task has no result due to a failed run."
            ) from self._exception
        return self._result

    @result.setter
    def result(self, _result):
        """Set task result."""
        self.status = TaskStatus.SUCCESS
        self._result = _result

    @property
    def exception(self):
        """Get task exception."""
        if self._result:
            raise ValueError("Task has no exception due to a successful run.")
        return self._exception

    @exception.setter
    def exception(self, _exception):
        """Set task exception."""
        self.status = TaskStatus.FAILED
        self._exception = _exception


class Task:
    """Task object maintains the running status."""

    def __init__(self, type, runner, name, payload):
        self.type = type
        self.runner = runner
        self.name = name
        self.payload = payload

    def __repr__(self):
        return (
            f"Task(type={self.type}, "
            f"name={self.name}, "
            f"payload={self.payload})"
        )

    def __str__(self):
        return f"task.{self.type}.{self.name}"

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, o):
        return self.type == o.type and self.name == o.name

    def should_run(self, context):
        return all(
            hcl2.evaluate(depends_on, context)
            for depends_on in self.payload.get(DEPENDS_ON_KEY, [])
        )

    def should_rerun(self, task, context):  # pylint: disable=too-many-branches
        if "_retry" not in self.payload:
            return task.run

        _retry = hcl2.evaluate(self.payload.get("_retry", {}), context)
        _retry_params = {}
        stop_after = _retry.get("stop_after", "1 times")
        stopers = []
        for _stop_after in stop_after.split("|"):
            _stop_after = _stop_after.strip()
            if not _stop_after:
                continue
            if "times" in _stop_after:
                attempts = int(_stop_after.replace("times", "").strip())
                stoper = stop_after_attempt(attempts)
            elif "seconds" in _stop_after:
                seconds = float(_stop_after.replace("seconds", "").strip())
                stoper = stop_after_delay(seconds)
            else:
                raise ValueError(f"invalid _retry.stop_after: {_stop_after}")
            stopers.append(stoper)

        _retry_params["stop"] = reduce(lambda a, b: a | b, stopers)
        _retry_params["wait"] = _retry.get("wait", None)
        _retry_params["reraise"] = True
        _retry_params["after"] = after_log(logger, logging.DEBUG)
        return retry(**_retry_params)(task.run)

    def eval_payload(self, context):
        return hcl2.evaluate(
            {k: v for k, v in self.payload.items() if not k.startswith("_")},
            context,
        )

    async def run(self, context):
        """Run a task."""
        if self.type == "hcl2_template":
            context = dict(
                context,
                **hcl2.evaluate(self.payload.get("context", {}), context),
            )

        if not self.should_run(context):
            logger.info(
                '"task.%s.%s" is canceled due to falsy deps.',
                self.type,
                self.name,
            )
            return TaskResult(TaskStatus.CANCELED)

        _payload = self.eval_payload(context)
        context["task"][self.type][self.name].update(_payload)

        task_result = TaskResult(TaskStatus.PENDING)

        try:
            logger.info('"task.%s.%s" is started.', self.type, self.name)
            task = self.runner(**_payload)
            task_run = self.should_rerun(task, context)
            task_result.result = (
                await task_run(context)
                if inspect.iscoroutinefunction(task.run)
                else await utils.to_thread(task_run, context)
            ) or {}
            task_result.result.update(_payload)
            logger.info('"task.%s.%s" is successful.', self.type, self.name)
        except Exception as err:  # pylint: disable=broad-except
            task_result.exception = err
            logger.info('"task.%s.%s" is failed.', self.type, self.name)
            traceback.print_exc()

        return task_result


class SequentialRunner:
    """This runner runs the flow tasks sequentially."""

    def __init__(self, flow):
        self.flow = flow
        self.results = {}

    def check_depends_on(self, task):
        for upstream_task in self.flow.graph.successors(task):
            if self.results[upstream_task].status in (
                TaskStatus.FAILED,
                TaskStatus.CANCELED,
            ):
                return False
        return True

    async def run_task(self, task, context):
        if not self.check_depends_on(task):
            logger.info(
                '"%s" is canceled due to previous task failed run.',
                f"task.{task.type}.{task.name}",
            )
            return TaskResult(TaskStatus.CANCELED)

        task_result = await task.run(context)

        if task_result.status == TaskStatus.SUCCESS:
            context["task"][task.type][task.name] = task_result.result

        return task_result

    async def run(self, context):
        """Run flow tasks."""
        for task in self.flow:
            self.results[task] = await self.run_task(task, context)


class Flow:
    """Flow object manages the flow graph and the order of task executions."""

    def __init__(self, name, runner_cls=None):
        self.name = name
        self.graph = networkx.DiGraph()
        self.runner = (runner_cls or SequentialRunner)(self)
        self.vars = {}
        self.functions = {}

    def __iter__(self):
        """Iterate through all tasks in a dependent order."""
        try:
            return reversed(list(networkx.topological_sort(self.graph)))
        except networkx.exception.NetworkXUnfeasible as err:
            raise RunflowAcyclicTasksError(str(err)) from err

    @property
    def exception(self):
        return next(
            (
                task_result.exception
                for task_result in self.runner.results.values()
                if task_result.status == TaskStatus.FAILED
            ),
            None,
        )

    @classmethod
    def from_spec(cls, source):
        """Load flow from a .hcl file content."""
        try:
            flow = hcl2.loads(source)
        except LarkError as err:
            raise RunflowSyntaxError(str(err)) from err

        assert "flow" in flow, "Need a flow block"
        assert len(flow["flow"]) == 1, "Runflow spec should have only one flow"

        flow_spec = flow["flow"][0]
        flow_name, flow_body = next(iter(flow_spec.items()))
        flow = cls(name=flow_name)
        flow.load_flow_spec_body(flow_body)
        return flow

    @classmethod
    def from_specfile(cls, path):
        """Load flow from a given file path."""
        with open(path) as file:
            flow_spec = file.read()
        return cls.from_spec(flow_spec)

    def add_task(self, task):
        """Add task to flow graph."""
        self.graph.add_node(task)

    def set_dependency(self, task, depends_on):
        """Let `task` depends on `depends_on`."""
        self.graph.add_edge(task, depends_on)

    def set_default_var(self, name, value):
        """Set default value for variable."""
        self.vars[name] = value

    def load_function(self, func_name, import_string):
        """Load the imported function to task func namespace."""
        function = utils.import_module(import_string)
        self.functions[func_name] = function

    # pylint: disable=no-self-use
    def load_flow_tasks_from_spec(self, tasks_spec):
        """Load the `task` blocks."""
        for task_spec in tasks_spec:
            for task_type in task_spec:
                task_class = get_task_class(task_type)
                for task_name in task_spec[task_type]:
                    task_payload = dict(task_spec[task_type][task_name])
                    yield Task(task_type, task_class, task_name, task_payload)

    def load_task_by_task_reference(self, depends_on):
        """Find task by a reference like `task.TASK_TYPE.TASK_NAME`."""
        task_key = depends_on.split(".")
        assert len(task_key) == 3 and task_key[0] == "task"

        _, task_type, task_name = task_key

        for task in self.graph.nodes:
            if task.type == task_type and task.name == task_name:
                return task

        raise RunflowSyntaxError(
            f"Task depends on {task_key} "
            f"but the dependent task does not exist"
        )

    def load_flow_tasks_dependencies(self, task):
        """Find task dependencies."""
        deps_set = set()
        for key, value in task.payload.items():
            hcl2.resolve_deps(key, deps_set)
            hcl2.resolve_deps(value, deps_set)

        for dep in deps_set:
            yield self.load_task_by_task_reference(dep)

    def set_tasks_dependencies(self):
        """Walk the task graph and sort out the task dependencies."""
        for task in self.graph.nodes:
            explicit_deps = self.load_flow_tasks_dependencies(task)
            for dep in explicit_deps:
                self.set_dependency(task, dep)

    def load_flow_variables(self, vars_spec):
        """Load the `variable` block."""
        for var_spec in vars_spec:
            var_name = next(iter(var_spec.keys()))
            var_value_spec = next(iter(var_spec.values()))
            var_value = config(
                f"RUNFLOW_VAR_{var_name}", default=None
            ) or var_value_spec.get("default", None)
            self.set_default_var(var_name, var_value)

    # pylint: disable=no-self-use
    def load_flow_imported_tasks(self, tasks):
        """Load the `import.tasks` attribute."""
        for task_name, task_impl in tasks.items():
            register_task_class(task_name, task_impl)

    def load_flow_imported_functions(self, functions):
        """Load the `import.functions` attribute."""
        for func_name, func_import in functions.items():
            self.load_function(func_name, func_import)

    def load_flow_imports(self, imports):
        """Load the `import` block."""
        for _import in imports:
            self.load_flow_imported_tasks(_import.get("tasks", []))
            self.load_flow_imported_functions(_import.get("functions", {}))

    def load_flow_tasks(self, tasks):
        for task in self.load_flow_tasks_from_spec(tasks):
            self.add_task(task)

    def load_flow_spec_body(self, spec):
        """Load the body of a flow block."""
        self.load_flow_imports(spec.get("import", []))
        self.load_flow_variables(spec.get("variable", []))
        self.load_flow_tasks(spec.get("task", []))
        self.set_tasks_dependencies()

    def make_run_context(self, vars=None):
        """Prepare the context for a task run."""
        context = {
            "var": dict(self.vars, **dict(vars or {})),
            "task": {},
            "func": self.functions,
        }
        for task in self:
            context["task"].setdefault(task.type, {})
            context["task"][task.type].setdefault(task.name, {})
        return context

    async def run(self, vars=None):
        """Run a flow."""
        context = self.make_run_context(vars)
        await self.runner.run(context)
        return context



from base64 import b64decode, b64encode

import fsspec


class FileWriteTask:
    def __init__(self, filename, content=None, b64content=None, fs=None):
        fs = fs if fs else {"protocol": "file"}
        self.filesystem = fsspec.filesystem(**fs)
        self.filename = filename
        if content is not None:
            self.content = content
        elif b64content is not None:
            self.content = b64decode(b64content)
        else:
            raise ValueError(
                "at least one of `content` / `b64content` must have value"
            )

    async def run(self, context):
        if self.filename == "/dev/stdout":
            print(str(self.content))
            return

        content = (
            self.content
            if isinstance(self.content, bytes)
            else self.content.encode("utf-8")
        )

        with self.filesystem.open(self.filename, "wb") as file:
            file.write(content)


class FileReadTask:
    def __init__(self, filename, fs=None):
        fs = fs if fs else {"protocol": "file"}
        self.filesystem = fsspec.filesystem(**fs)
        self.filename = filename

    async def run(self, contenxt):
        with self.filesystem.open(self.filename, "rb") as file:
            content = file.read()
            return {
                "content": content.decode("utf-8"),
                "b64content": b64encode(content).decode("utf-8"),
            }



"""Run a flow spec."""

from .core import Flow
from .utils import import_module, run_async


def loadflow(path=None, source=None, module=None, flow=None):
    """Load a flow object."""
    if path:
        flow = Flow.from_specfile(path)
    elif source:
        flow = Flow.from_spec(source)
    elif module:
        flow = import_module(module)
    return flow


async def runflow_async(
    path=None, source=None, module=None, flow=None, vars=None
):
    """Run a flow object (async)."""
    _flow = loadflow(path=path, source=source, module=module, flow=flow)
    assert _flow and isinstance(_flow, Flow)
    await _flow.run(vars or {})


def runflow(path=None, source=None, module=None, flow=None, vars=None):
    """Run a flow object (sync)."""
    run_async(
        runflow_async(
            path=path, source=source, module=module, flow=flow, vars=vars
        )
    )



from runflow import loadflow
from runflow.hcl2 import loads, evaluate


class FlowRunTask:
    def __init__(
        self, path=None, source=None, module=None, vars=None, export=None
    ):
        self.path = path
        self.source = source
        self.module = module
        self.vars = vars or {}
        self.export = export or []

    async def run(self, context):
        flow = loadflow(path=self.path, source=self.source, module=self.module)
        flow_context = await flow.run(vars=self.vars)

        if flow.exception:
            raise flow.exception

        return {
            key: evaluate(loads(value, "eval"), flow_context)
            for export in self.export
            for key, value in export.items()
        }



from .errors import RunflowReadOnlyError
from .utils import import_module


task_implementations = {
    "flow_run": {
        "class": "runflow.contribs.flow:FlowRunTask",
    },
    "bash_run": {
        "class": "runflow.contribs.bash:BashRunTask",
    },
    "docker_run": {
        "class": "runflow.contribs.docker:DockerRunTask",
    },
    "file_read": {
        "class": "runflow.contribs.local_file:FileReadTask",
    },
    "file_write": {
        "class": "runflow.contribs.local_file:FileWriteTask",
    },
    "hcl2_template": {
        "class": "runflow.contribs.template:Hcl2TemplateTask",
    },
    "http_request": {
        "class": "runflow.contribs.http:HttpRequestTask",
    },
    "sql_exec": {
        "class": "runflow.contribs.sql:SqlExecTask",
    },
    "sql_row": {
        "class": "runflow.contribs.sql:SqlRowTask",
    },
}


class ImmutableDict(dict):
    """Dict but immutable.

    Getitem is allowed; set/del/clear are disallowed.
    """

    def __init__(self, target):
        super().__init__()
        self.target = target

    def __getitem__(self, item):
        return self.target[item]

    def __delitem__(self, key):
        raise RunflowReadOnlyError

    def __setitem__(self, key, value):
        raise RunflowReadOnlyError

    def clear(self):
        raise RunflowReadOnlyError

    def __contains__(self, item):
        return item in self.target

    def __iter__(self):
        yield from self.target


_registry = {}

registry = ImmutableDict(_registry)


def register_task_class(name, cls, overwrite=False):
    """Register a task class with a name."""
    if isinstance(cls, str):
        if name in task_implementations and not overwrite:
            raise ValueError(
                f"Task name {name} has been taken in the "
                "default task implementation."
            )
        task_implementations[name] = {"class": cls}
        return
    if name in registry and not overwrite:
        raise ValueError(f"Task name {name} has been taken in the registry.")
    if not hasattr(cls, "run"):
        raise ValueError("cls is not runnable.")
    _registry[name] = cls


def get_task_class(name):
    """Get the task class by name."""
    if name not in registry:
        if name not in task_implementations:
            raise ValueError(f"Unknown task type: {name}")
        impl = import_module(task_implementations[name]["class"])
        register_task_class(name, impl)
    return registry[name]



try:
    import docker
except ImportError:
    print("Please install Python package `docker` to use `docker_run` task.")
    docker = None

from runflow.utils import to_thread


class DockerRunTask:
    def __init__(self, image, command=None, **kwargs):
        self.image = image
        self.command = command
        self.args = kwargs
        self.client = docker.from_env()

    async def run(self, context):
        stdout = await to_thread(
            self.client.containers.run,
            image=self.image,
            command=self.command,
            detach=False,
            remove=True,
            **self.args
        )
        return {
            "stdout": stdout.decode("utf-8").strip(),
        }



import asyncio

from runflow.errors import RunflowTaskError


class BashRunTask:
    def __init__(self, command, env=None):
        self.command = command
        env = env or {}
        self.env = {k: str(v) for k, v in env.items()}

    async def run(self, context):
        proc = await asyncio.create_subprocess_shell(
            self.command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=self.env,
        )

        stdout, stderr = await proc.communicate()
        stdout = stdout.decode("utf-8").strip()
        stderr = stderr.decode("utf-8").strip()

        if stdout:
            print(stdout)

        if proc.returncode == 0:
            return dict(
                returncode=proc.returncode,
                stdout=stdout,
                stderr=stderr,
            )

        raise RunflowTaskError(
            dict(
                returncode=proc.returncode,
                stdout=stdout,
                stderr=stderr,
            )
        )
